/*
 *  Program to compute shading and shadows for terrain data.
 *  Stephanie Weirich
 *
 */
#include "matrix.h"
#include "shade.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

/****** Global Variables *********/
int nverts;               /* Number of vertices in world */
int npolys;               /* Number of polygons */
Poly_vert* vlist;         /* array of vertices */
Triangle* plist;          /* array of polygons (with references to vertices) */
tree* root;
Window maximum;

double EastEnd;
double WestEnd;
double NorthPole;
double SouthPole;
double MtHimalaya = 0;
double DeathValley;


/* Parameters to the program */
double angle = M_PI/4;    
double red_diff = 0.1;
double blue_diff = 0.1;
double green_diff = 0.1;
double red_amb = 0.2;
double blue_amb = 0.2;
double green_amb = 0.2;
double epsilon = 0.0;
double xfactor = 10.0;
double yfactor = 1.0;
double zfactor = 1.0;
file_t filetype = ASCII;
char infilename[50] = {"infile.dat"};
char colorfilename[50] = {'\0'};
char outfilename[50] = {"rgb.out"};

/******************************
This function parses the command 
line flags as described in the 
file README
********************************/

#ifdef ANSI 
void parseCommandLine(int argc, char** argv)
#else
void parseCommandLine(argc, argv)
     int argc;
     char ** argv;
#endif
{
  int i=1;
  if (argc == 2)
    {
      strcpy(infilename, argv[1]);
      return;
    }
  while (i<argc)
    {
      switch (argv[i][1])
	{            
	  /** look at first letter of flag **/
	case 's': angle = atof(argv[i+1]); break;
	case 'i': strcpy(infilename, argv[i+1]); break;
	case 'b': filetype = BINARY; i--; break;
	case 'c': strcpy(colorfilename, argv[i+1]); break;
	case 'o': strcpy(outfilename,argv[i+1]); break;
	case 'd':
	  if (argv[i][2] == 'r')
	    {
	      red_diff = atof(argv[i+1]);
	      break;
	    }
	  if (argv[i][2] == 'g')
	    {
	      green_diff = atof(argv[i+1]);
	      break;
	    }
	  if (argv[i][2] == 'b')
	    {
	      blue_diff = atof(argv[i+1]);
	      break;
	    }
	  myerror(3);
	  break;
	case 'a':
	  if (argv[i][2] == 'r')
	    {
	      red_amb = atof(argv[i+1]);
	      break;
	    }
	  if (argv[i][2] == 'g')
	    {
	      green_amb = atof(argv[i+1]);
	      break;
	    }
	  if (argv[i][2] == 'b')
	    {
	      blue_amb = atof(argv[i+1]);
	      break;
	    }
	  myerror(3);
	  break;
	case 't': epsilon = atof (argv[i+1]);break;
	case 'x': xfactor = atof (argv[i+1]);break;
	case 'y': yfactor = atof (argv[i+1]);break;
	case 'z': zfactor = atof (argv[i+1]);break;
	default: myerror(3);
	}
      i+=2;
    }
}


/* opens file and creates data structure, calculating important 
 * initial variables and adjusting the data.
 */
void readInData()
{
  int i,j;
  double** Vertices; /* Temporary storage for reading in data */
  int** Polygons;
  int ncolors;

  if (filetype == BINARY)
    Read_binary_polygon_data(infilename, &nverts, &Vertices, 
			     &npolys, &Polygons);
  else
    Read_ascii_polygon_data(infilename, &nverts, &Vertices, 
			    &npolys, &Polygons);
  
  vlist=(Poly_vert*)malloc(nverts*sizeof(Poly_vert));
  if (!vlist)
    myerror(2);
  for(i=0; i<nverts; i++) 
    {
      vlist[i].x = Vertices[i][0];
      vlist[i].y = Vertices[i][1];
      vlist[i].z = Vertices[i][2];
      
      /**** Compensate so triangles are big enough *******/
      vlist[i].x *=xfactor;
      vlist[i].y *=yfactor;
      vlist[i].z *=zfactor;
      
      /******** Find Landmarks *********/
      if (i==0 || (vlist[i].x > EastEnd))
	EastEnd = vlist[i].x;
      if (i==0 || (vlist[i].x < WestEnd))
	WestEnd = vlist[i].x;
      if (i==0 || vlist[i].z > NorthPole)
	NorthPole = vlist[i].z;
      if (i==0 || vlist[i].z < SouthPole)
	SouthPole = vlist[i].z;
      if (vlist[i].y > MtHimalaya)
	MtHimalaya  = vlist[i].y;
      if (i==0 || vlist[i].y < DeathValley)
	  DeathValley = vlist[i].y;
      
      /******* Define initial colors ******/
      if (vlist[i].y == 0)         /* Blue Water */
	{
	  vlist[i].r = 0.0;
	  vlist[i].g = 0.0;
	  vlist[i].b = 1.0;
	}
      else                        /* Green Land */
	{
	  vlist[i].r = 0.0;
	  vlist[i].g = 1.0;
	  vlist[i].b = 0.0;
	}
      for(j=0; j<4; j++)           /* initialize */
	vlist[i].normal[j] = 0.0;
      vlist[i].shaded = FALSE;
    }

  /******** center vertices ************/
  for (i = 0; i<nverts; i++)
    {
      vlist[i].x -= WestEnd;
      vlist[i].z -= SouthPole;
      vlist[i].y -= MtHimalaya;
    }
  
  /******* Polygons! ***********/
  plist =(Triangle*)malloc(npolys*sizeof(Triangle));
  if (!plist)
    myerror(2);
  for (i=0; i<npolys; i++)
    { 
      plist[i].v[0]=Polygons[i][0];
      plist[i].v[1]=Polygons[i][1];
      plist[i].v[2]=Polygons[i][2];
      plist[i].n =3;
      plist[i].degenerate = FALSE;
    }
  
  for (i=0; i<nverts; i++)
    free(Vertices[i]);
  free(Vertices);
  for (i=0; i<npolys; i++)
    free(Polygons[i]);
  free(Polygons);


  /**** Read colors from file if there is one ********/
  if (colorfilename[0] != '\0')
    {
      if (filetype == BINARY)
	Read_binary_color(colorfilename, &ncolors, Vertices);
      else
	Read_ascii_color(colorfilename, &ncolors, Vertices);
      if (ncolors != nverts)
	myerror(4);
      for (i=0; i< nverts; i++)
	{
	  vlist[i].r = Vertices[i][0];
	  vlist[i].g = Vertices[i][1];
	  vlist[i].b = Vertices[i][2];
	}
      for (i=0; i<nverts; i++)
	free(Vertices[i]);
      free(Vertices);
    }
}


/*********************************************
  Average's the normal of the polygon into the 
  normal of the vector 
**********************************************/
#ifdef ANSI
void averageIn(vector n, Poly_vert* v)
#else
     averageIn(n,v)
     vector n;
     Poly_vert* v;
#endif    
{
  int i;
  for(i=0; i<3; i++)
      v->normal[i]+=n[i];
}

/*** Returns True if same, FALSE is different ****/
#ifdef ANSI
int cmpPoints(point* p0, point* p1)
#else
     int cmpPoints(p0, p1)
     point *p0, *p1;
#endif
{
  return ((p0->x == p1->x) && (p0->y == p1->y) && (p0->z == p1->z));
}

/*****************************************
  Averages the normals of all the polygons 
  into the normals of all the vertices.
  ****************************************/
void computeNormals()
{
  int i,j;
  vector n, a, b;
  int v1, v2, v0;
  point p1, p2, p0;

  for (i=0; i<npolys; i++)
    {
      v0 = plist[i].v[0];
      v1 = plist[i].v[1];
      v2 = plist[i].v[2];
      point_copy(&p0, &(vlist[v0]));
      point_copy(&p1, &(vlist[v1]));
      point_copy(&p2, &(vlist[v2]));
      if (cmpPoints(&p0, &p1) || cmpPoints(&p1, &p2) || 
	  cmpPoints(&p0, &p2))
	{
	  plist[i].degenerate = TRUE;
	  printf("Degenerate polygon %d\n", i);
	}
      else
	{
	  unit_line_vector(a, &p1, &p0);
	  unit_line_vector(b, &p2, &p0);
	  cross_product(n, a, b);
	  averageIn(n, &vlist[v0]);
	  averageIn(n, &vlist[v1]);
	  averageIn(n, &vlist[v2]);
	  for (j=0; j<4; j++)
	    plist[i].normal[j] = n[j];  
	}
    }
  for (i=0; i<nverts; i++)
    unit_vector(vlist[i].normal, vlist[i].normal);
}

#ifdef ANSI
void vert2vector(vector vect, Poly_vert* vert)
#else 
     void  vert2vector(vect, vert)
     vector vect;
     Poly_vert* vert;
#endif
{
  vect[0] = vert->x;
  vect[1] = vert->y;
  vect[2] = vert->z;
  vect[3] = 1.0;
}

#ifdef ANSI
double max(double x, double y)
#else
     double max(x, y)
     double x;
     double y;
#endif
{
  return ((x > y)? x : y);
}
/**************************************
  Creates a bounding box for triangle p
  and returns it in win
  *************************************/

#ifdef ANSI
void computeLimits(Window* win,  Triangle* p)
#else 
     void computeLimits(win, p)
     Window* win;
     Triangle* p;
#endif
{
  double xmax = vlist[p->v[0]].sx;
  double xmin = xmax;
  double zmax = vlist[p->v[0]].sz;
  double zmin = zmax;
  int i;
  double currx, currz;

  for (i=1; i<3; i++)
    {
      currx = vlist[p->v[i]].sx;
      if (xmax < currx)
	xmax = currx;
      else
	if (xmin > currx)
	  xmin = currx;
      currz= vlist[p->v[i]].sz;
      if (zmax < currz)
	zmax = currz;
      else
	if (zmin > currz)
	  zmin = currz;
    }
  win->x0 = ceil(xmin-0.5);
  win->x1 = ceil(xmax-0.5);
  win->y0 = ceil(zmin-0.5);
  win->y1 = ceil(zmax-0.5);
}

/*************************************
  This function is called at every pixel
  in the scan-conversion.
  ***********************************/
#ifdef ANSI
pixelproc(int x, int z, Poly_vert* pnt, Y_Struct** buffer, Window* win)
#else
     pixelproc(x, z, pnt, buffer, win)
     int x, z;
     Poly_vert *pnt;
     Y_struct** buffer;
     Window* win;
#endif
{
  int my_x, my_y;
  my_x = x - win->x0;
  my_y = z - win->y0;
  if (pnt->sy > buffer[my_x][my_y].y) 
    buffer[my_x][my_y].y = pnt->sy;
}

/*  Computes the shadows on the scene by first building a tree of 
 *  all the bounding boxes of the back facing polygons (since only
 *  have to scan where there are back facing polygons), then 
 *  allocating the Z-buffers in that tree, and finally scanning
 *  all of the polygons, into those Z-buffers.
 */
void castShadows()
{
  long int totarea, buffarea;
  int i,j;
  matrix transform;
  vector real, sky;
  vector temp;
  Window* win;
  Window otherwin;
  vector topright, botleft, botright;  /* Used to calculate largest extent */
  double maxsize, altitude;            /* of Z-buffer.                     */
  int xsize, ysize;
  double xtemp, twotemp;
  tree* root = NULL;                   /* Root of Z-buffer tree            */

  /**** transform set of vertices *****/
  id_matrix(transform);
  rotate_z(transform, angle);
  for (i=0; i<nverts; i++)
    {
      vert2vector(real, &vlist[i]);
      vector_transform(sky, real, transform);
      vlist[i].sx = sky[0];
      vlist[i].sy = sky[1];
      vlist[i].sz = sky[2];
      vlist[i].sw = sky[3];
    }
  /********** Find max size for Window *************/
  altitude = DeathValley - MtHimalaya;

  double2vector(topright, (EastEnd-WestEnd), 0, 0);
  double2vector(botright, (EastEnd-WestEnd), altitude, 0);
  double2vector(botleft, 0, altitude, 0);
  
  vector_transform(topright, topright, transform);
  vector_transform(botright, botright, transform);
  vector_transform(botleft, botleft, transform);

  maxsize = max(botleft[0],  max(topright[0], botright[0]));
  xsize = (int)ceil(maxsize) +1;
  ysize = (int)ceil(NorthPole - SouthPole)+1;

  totarea = (long)xsize*ysize;
  /*** Want next power of two *****/
  xtemp  = log((double)xsize);
  twotemp = log((double)2.);
  xsize = 1 << ((int)(xtemp/twotemp)+1);
  ysize = 1 << ((int)(log((double)ysize)/twotemp)+1);
  maximum.x0 = 0;
  maximum.x1 = xsize;
  maximum.y0 = 0;
  maximum.y1 = ysize;
  /********* Build Scan Tree *************/
  for (i=0; i<npolys; i++)
    {
      if (!plist[i].degenerate)
	{
	  vector_transform(temp, plist[i].normal, transform);
	  if (temp[1] <= 0)
	    { 
	      win = (Window*)malloc(sizeof(Window));
	      computeLimits(win,&plist[i]);
	      addToTree(root, win, &maximum, 0);
	    }
	}
    }
  buffarea = area(root);
  printf("Total area: %ld, buff %ld\n", totarea, buffarea);
  printf("Percent of window scanned %f\n", ((float)buffarea/(float)totarea)*100);
  allocateBuffer(root);
  for (i=0; i< npolys; i++)
    {
      if (!plist[i].degenerate)
	{
	  vector_transform(temp, plist[i].normal, transform);
	  if (temp[1] >= 0) 
	    {
	      computeLimits(&otherwin,&plist[i]);
	      ScanTree(root, &otherwin, &maximum, &plist[i], vlist, 0);
	    }
	}
    }
}

/* This function, for every vertex, checks to see if it is in shadow, and
   if it isn't adds in the diffuse quality of the light to the vertex 
   intensity. */

void computeIntensities()
{
  double *red_ints; /* Temporary intensity for each vertex */
  double *blue_ints;
  double *green_ints;
  double dot;
  int  i,j;
  point p0, p1;
  vector lightnormal;
  int x, z;

  red_ints = (double*)malloc(nverts*sizeof(double));
  blue_ints = (double*)malloc(nverts*sizeof(double));
  green_ints = (double*)malloc(nverts*sizeof(double));
  
  /*** Find Normal of light ***/
  double2point(&p0, 0.0, 0.0, 0.0);
  double2point(&p1,  sin(angle), cos(angle), 0.0);
  unit_line_vector(lightnormal, &p1, &p0);
  
  for (i = 0; i < nverts; i++)
    {
      x = ceil(vlist[i].sx-0.5);
      z = ceil(vlist[i].sz-0.5);
      red_ints[i] = 0;
      green_ints[i] = 0;
      blue_ints[i] = 0;
      if (!InShadow(root, x, z, vlist[i].sy, &maximum, 0))
	{
	  dot_product(&dot, lightnormal, vlist[i].normal);
	  if (dot < 0)
	    dot = 0.0;
	  red_ints[i] = red_diff * dot;
	  blue_ints[i] = blue_diff * dot;
	  green_ints[i] = green_diff * dot;
	}
      red_ints[i] += red_amb;
      blue_ints[i] += blue_amb;
      green_ints[i] += green_amb;
      
      if (red_ints[i] > 1.0)
	red_ints[i] =1.0;
      else
	if (red_ints[i] < 0.0)
	  red_ints[i] =0.0;
      
      if (blue_ints[i] >1.0)
	blue_ints[i] = 1.0;
      else
	if (blue_ints[i] < 0.0)
	  blue_ints[i] = 0.0;
      
      if (green_ints[i] > 1.0)
	green_ints[i] = 1.0;
      else
	if (green_ints[i] < 0.0)
	  green_ints[i] = 0.0;
      
      vlist[i].r *= red_ints[i];
      vlist[i].g *= green_ints[i];
      vlist[i].b *= blue_ints[i];
    }
  free(red_ints);
  free(blue_ints);
  free(green_ints);
}

/* prints out the result in a color file. (number of vertices, followed by 
   the rgb values for each vertex.  */      
void printResults()
{
  FILE *out;
  int i;
  double * buffer;
 
  out = fopen(outfilename, "w");
  if (!out)
    myerror(0);

  if (filetype == BINARY)
    {
      buffer = (double *) malloc(3*nverts*sizeof(double));
      for (i = 0; i< nverts; i+=3)
	{
	  buffer[i] = vlist[i/3].r;
	  buffer[i+1] = vlist[i/3].g;
	  buffer[i+2] = vlist[i/3].b;
	}
      fwrite(&nverts, sizeof(int), 1, out);
      fwrite(buffer,sizeof(double),3*nverts, out);
    }
  else
    {
      fprintf(out, "%d\n", nverts);
      for (i =0; i<nverts; i++)
	fprintf(out, "%lf %lf %lf \n", vlist[i].r, vlist[i].g, vlist[i].b);
    }
  fflush(out);
  fclose(out);
}
      
	    
#ifdef ANSI
void main(int argc, char** argv)
#else
     void main(argc, argv)
     int argc;
     char** argv;
#endif
{
  int i;
  printf("Parsing command line\n");
  parseCommandLine(argc, argv);
  printf("Reading data\n");
  readInData();
  for(angle=0; angle<1.5; angle+=0.1)
    {
      printf("Angle = %lf\n", angle);
      /* printf("Computing normals\n"); */ 
      computeNormals();
      /* printf("Casting shadows\n"); */
      castShadows();
    }
/*      printf("Calculating vertex intensities\n");
      computeIntensities();
  printf("Writing results\n");
  printResults();  */
}


